/***************************************************************************
                     cdf.cpp  -  netCDF file data source
                             -------------------
    begin                : 17/06/2004
    copyright            : (C) 2004 Nicolas Brisset <nicodev@users.sourceforge.net>
    email                : kst@kde.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "netcdf_source.h" // Local header for the kst netCDF datasource

#include <kdebug.h>

#include <qfile.h>
#include <qfileinfo.h>

#include <ctype.h>
#include <stdlib.h>

NetcdfSource::NetcdfSource(KConfig *cfg, const QString& filename, const QString& type)
: KstDataSource(cfg, filename, type) {
    if (!type.isEmpty() && type != "netCDF") {
      return;
    }

    _valid = false;
    _maxFrameCount = 0;
    
    _filename = filename;
    if (initFile()) {
      _valid = true;
    }
}


NetcdfSource::~NetcdfSource() {
  _maxFrameCount = 0;
  delete _ncfile;
}


bool NetcdfSource::initFile() {
  kdDebug() << "netCDF datasource, " << _filename << ": entering initFile()" << endl;
  _ncfile = new NcFile(_filename.latin1());
  if (!_ncfile) {
    kdDebug() << _filename << ": failed to open in initFile()" << endl;
    return false;
  }

  // Query field list and store it in _fieldList
  kdDebug() << _filename << ": building field list" << endl;
  _fieldList.clear();
  // I wonder what happens with INDEX when variables do not all have the same length
  _fieldList += "INDEX";
  // Add scalar variables only !
  NcVar *var;
  int nb_vars = _ncfile->num_vars();
  kdDebug() << nb_vars << " vars found in total" << endl;
  for (int i = 0; i < nb_vars; i++) {
    var = _ncfile->get_var(i);
    if (var->num_dims()==1) {
      _fieldList += var->name();
      _frameCounts[var->name()] = var->num_vals();
      if ( var->num_vals() > _maxFrameCount) _maxFrameCount = var->num_vals();
      kdDebug() << var->name() << ": " << var->num_vals() << " values" << endl;
    }
  }
  kdDebug() << "of which " << _fieldList.count() << " scalar variables"  << endl;
  kdDebug() << "Max samples: " << _maxFrameCount << endl;
  return true;
}


KstObject::UpdateType NetcdfSource::update(int u) {
  Q_UNUSED(u)
  // It probably would be easy to allow for "growing" netCDF files by requerying dimension sizes here
  // but it can be added later in case someone needs it. We'd also need to make a function call in frameCount()
  // instead of caching the values in initFile().
  // For now, just return ::NO_CHANGE 
  return KstObject::NO_CHANGE;
}


int NetcdfSource::readField(double *v, const QString& field, int s, int n) {
  int i;
  NcType dataType = ncNoType; /* netCDF data type */
  NcVar *var;                 /* netCDF variable */
  long maxRec;                /* max number of values for this var */
  size_t start = s;
  size_t count = n;
  kdDebug() << "Entering NetcdfSource::readField with params: " << field << ", from " << s << " for " << n << " values" << endl;

  // Handle the special case where we query INDEX
  if (field.lower() == "index") {
    if (n < 0) {
      v[0] = double(s);
      return 1;
    }
    for (int i = 0; i < n; ++i) {
      v[i] = double(s + i);
    }
    return n;
  }

  // If n<0 set it to 1 as suggested by George Staikos
  // (needs to be documented better I guess !)
  if (n < 0) {
    n = 1;
  }

  // If not INDEX, look into the netCDF file...
  // Variable selection
  var = _ncfile->get_var(field.latin1());
  if (!var) {
    return -1;
    kdDebug() << "Queried field " << field << " which can't be read" << endl;
  }
  // Number of samples to avoid reading beyond
  maxRec = var->num_vals();
  // Type
  dataType = var->type();
  
  switch (dataType) {
    case ncShort:
      {
        short *temp_short = new short[n];
        if (nc_get_vara_short(_ncfile->id(), var->id(), &start, &count, temp_short)) return -1;
        for (i = 0; i < n; i++) {
          v[i] = temp_short[i];
        }
      }	
      break;
    case ncInt:
      {
        int *temp_int = new int[n];
        if (nc_get_vara_int(_ncfile->id(), var->id(), &start, &count, temp_int)) return -1;
        for (i = 0; i < n; i++) {
          v[i] = temp_int[i];
        }      
      }
      break;
    case ncFloat:
      {
        float *temp_float = new float[n];
        if (nc_get_vara_float(_ncfile->id(), var->id(), &start, &count, temp_float)) return -1;
        for (i = 0; i < n; i++) {
          v[i] = temp_float[i];
        } 
      } 
      break;
    case ncDouble: 
      if (nc_get_vara_double(_ncfile->id(), var->id(), &start, &count, v)) return -1;
      break;
    default:
      kdDebug() << field << ": wrong datatype for kst, no values read" << endl;
      return -1;
      break;
  }
  // Uncomment following to see that it DOES read the values successfully
  //kdDebug() << field << "[" << i << "]=" << v[i] << endl;

  kdDebug() << "Finished reading " << field << endl;
  
  return n;
}


bool NetcdfSource::isValidField(const QString& field) const {  
  return _fieldList.contains(field);
}


int NetcdfSource::samplesPerFrame(const QString &field) {
  Q_UNUSED(field)
  return 1; 
}


int NetcdfSource::frameCount(const QString& field) const {
  // Handle the case where the argument is null (return max number of records for all vars)
  if (field.isEmpty() || field == "INDEX") {
    return _maxFrameCount;
  }
  // Other case : count queried specifically for one field
  else { 
    return _frameCounts[field];
  }
}


QString NetcdfSource::fileType() const {
  return "netCDF";
}


void NetcdfSource::save(QTextStream &ts, const QString& indent) {
  // FIXME (copied from ascii.cpp !)
  KstDataSource::save(ts, indent);
}


bool NetcdfSource::isEmpty() const {
  return frameCount() < 1;
}


extern "C" {
KstDataSource *create_netcdf(KConfig *cfg, const QString& filename, const QString& type) {
  return new NetcdfSource(cfg, filename, type);
}

QStringList provides_netcdf() {
  QStringList rc;
  rc += "netCDF";
  return rc;
}

/** understands_netcdf: returns true if:
    - the file is readable (!)
    - the file can be opened by the netcdf library **/
int understands_netcdf(KConfig*, const QString& filename) {
  QFile f(filename);

  if (!f.open(IO_ReadOnly)) {
    kdDebug() << "Unable to read file !" << endl;
    return 0;
  }
  
  NcFile *ncfile = new NcFile(filename.latin1());
  if (ncfile->is_valid()) {
    kdDebug() << filename << " looks like netCDF !" << endl;
    delete ncfile;
    return 80;
  }
  else return 0;

}

}

// vim: ts=2 sw=2 et
